"""Web playground for Eryx."""

import io
import json
import re
import uuid
from contextlib import redirect_stdout

from flask import Flask, jsonify, render_template, request

from eryx.__init__ import CURRENT_VERSION
from eryx.frontend.lexer import tokenize
from eryx.frontend.parser import Parser
from eryx.runtime.environment import Environment
from eryx.runtime.interpreter import evaluate
from eryx.utils.pretty_print import pprint

app = Flask(__name__)
app.config["SEND_FILE_MAX_AGE_DEFAULT"] = 0
parser = Parser()
environments = {}

# https://stackoverflow.com/questions/19212665/python-converting-ansi-color-codes-to-html
COLOR_DICT = {
    "31": [(255, 105, 180), (128, 105, 180)],
    "32": [(57, 255, 20), (57, 128, 20)],
    "33": [(255, 193, 7), (128, 193, 7)],
    "34": [(0, 255, 255), (0, 128, 128)],
    "35": [(255, 105, 180), (128, 105, 180)],
    "36": [(0, 255, 255), (0, 128, 128)],
    "37": [(255, 255, 255), (192, 192, 192)],
}

COLOR_REGEX = re.compile(r"\[(?P<arg_1>\d+)(;(?P<arg_2>\d+)(;(?P<arg_3>\d+))?)?m")

BOLD_TEMPLATE = '<span style="color: rgb{}; font-weight: bolder">'
LIGHT_TEMPLATE = '<span style="color: rgb{}">'


def ansi_to_html(text):
    """Format ANSI color codes to HTML."""
    text = text.replace("[m", "</span>")

    def single_sub(match):
        argsdict = match.groupdict()
        if argsdict["arg_3"] is None:
            if argsdict["arg_2"] is None:
                color, bold = argsdict["arg_1"], 0
            else:
                color, bold = argsdict["arg_1"], int(argsdict["arg_2"])
        else:
            color, bold = argsdict["arg_2"], int(argsdict["arg_3"])

        if bold:
            return BOLD_TEMPLATE.format(COLOR_DICT[color][1])
        return LIGHT_TEMPLATE.format(COLOR_DICT[color][0])

    return COLOR_REGEX.sub(single_sub, text)


# ========


class TokenList:
    """List of tokens to use with the pretty printer."""

    def __init__(self, tokens: list):
        self.tokens = tokens


def get_unique_uuid(dictionary):
    """Get a unique UUID that does not exist in dictionary."""
    while True:
        uid = uuid.uuid4().hex
        if uid not in dictionary:
            return uid


@app.route("/")
def index():
    """Main page."""
    return render_template("index.html", version=CURRENT_VERSION)


@app.route("/ast", methods=["POST"])
def ast():
    """Ast route."""
    source_code = request.get_json()["source_code"]
    try:
        ast_nodes = parser.produce_ast(source_code)
    except RuntimeError as e:
        return jsonify({"error": str(e)})
    return jsonify(
        {
            "result": ansi_to_html(pprint(ast_nodes, print_output=False))
            .replace("\n", "<br>")
            .replace("\u001b", "")
        }
    )


@app.route("/result", methods=["POST"])
def result_route():
    """Result route."""
    request_json = request.get_json()
    source_code = request_json["source_code"]
    try:
        ast_nodes = parser.produce_ast(source_code)
    except RuntimeError as e:
        return jsonify({"error": str(e)})
    try:
        env = None
        if "env_uuid" in request_json:
            env_uuid = request_json["env_uuid"]
            env = environments.get(env_uuid)
        env = env if env else Environment()
        result = evaluate(ast_nodes, env)
    except RuntimeError as e:
        return jsonify({"error": str(e)})
    return jsonify(
        {
            "result": ansi_to_html(pprint(result, print_output=False))
            .replace("\n", "<br>")
            .replace("\u001b", "")
        }
    )


@app.route("/run", methods=["POST"])
def run():
    """Run route."""
    request_json = request.get_json()
    source_code = request_json["source_code"]
    try:
        ast_nodes = parser.produce_ast(source_code)
    except RuntimeError as e:
        return jsonify({"error": str(e)})
    try:
        env = None
        if "env_uuid" in request_json:
            env_uuid = request_json["env_uuid"]
            env = environments.get(env_uuid)
        env = env if env else Environment()
        output_buffer = io.StringIO()
        with redirect_stdout(output_buffer):
            evaluate(ast_nodes, env)
        captured_output = output_buffer.getvalue()
    except RuntimeError as e:
        return jsonify({"error": str(e)})
    return jsonify({"result": ansi_to_html(captured_output).replace("\n", "<br>")})


@app.route("/tokenize", methods=["POST"])
def tokenize_route():
    """Run route."""
    request_json = request.get_json()
    source_code = request_json["source_code"]
    try:
        tokens = tokenize(source_code)
    except RuntimeError as e:
        return jsonify({"error": str(e)})
    return jsonify(
        {
            "result": ansi_to_html(pprint(TokenList(tokens), print_output=False))
            .replace("\n", "<br>")
            .replace("\u001b", "")
        }
    )


@app.route("/repl", methods=["POST"])
def repl():
    """REPL route."""
    try:
        request_json = request.get_json()
    except json.JSONDecodeError:
        return jsonify({"error": "Invalid JSON"})

    if not request_json:
        return jsonify({"error": "No JSON provided"})

    if "action" not in request_json:
        return jsonify({"error": "No action provided"})

    action = request_json["action"]

    if action == "token":
        environment = Environment()
        env_uuid = get_unique_uuid(environments)
        environments[env_uuid] = environment
        return jsonify({"env_uuid": env_uuid})

    if action == "delete":
        if "env_uuid" not in request_json:
            return jsonify({"error": "No environment UUID provided"})

        env_uuid = request_json["env_uuid"]
        if env_uuid in environments:
            del environments[env_uuid]
        return jsonify({})

    return jsonify({"error": "Invalid action"})


@app.route("/static/<path:path>", methods=["GET"])
def static_route(path):
    """Static file route."""
    return app.send_static_file(path)


@app.route("/favicon.ico")
def favicon():
    """Serve the favicon."""
    return app.send_static_file("favicon.ico")


def start_playground(host="0.0.0.0", port=80):
    """Start the web playground."""
    app.run(host=host, port=port, debug=False, use_reloader=False)


if __name__ == "__main__":
    start_playground()



"""Environment class for storing variables (also called scope)."""

import sys
import time

from eryx.runtime.values import (
    ArrayValue,
    BooleanValue,
    FunctionValue,
    NativeFunctionValue,
    NullValue,
    NumberValue,
    ObjectValue,
    RuntimeValue,
    StringValue,
)


class Environment:
    """Environment class."""

    def __init__(self, parent_env: "Environment" = None):
        self.is_global = parent_env is None
        self.parent = parent_env
        self.constants = []
        self.variables = {}

        if self.is_global:
            self.setup_scope()

    def declare_variable(
        self, variable_name: str, value: RuntimeValue, constant: bool = False
    ) -> RuntimeValue:
        """Declare a variable in the current scope."""
        # Raise an exception if the variable is already declared
        if variable_name in self.variables:
            raise RuntimeError(f'Variable "{variable_name}" already declared')

        self.variables[variable_name] = value

        if constant:
            self.constants.append(variable_name)

        return value

    def assign_variable(self, variable_name: str, value: RuntimeValue) -> RuntimeValue:
        """Assign a value to a variable in the current scope."""
        environment = self.resolve(variable_name)

        if variable_name in environment.constants:
            raise RuntimeError(f'Cannot assign to constant variable "{variable_name}"')

        environment.variables[variable_name] = value
        return value

    def lookup_variable(self, variable_name: str) -> RuntimeValue:
        """Lookup a variable in the current scope."""
        environment = self.resolve(variable_name)
        return environment.variables[variable_name]

    def resolve(self, variable_name: str) -> "Environment":
        """Resolve a variable name to an environment."""
        # Return self if variable_name exists in the current scope
        if variable_name in self.variables:
            return self
        # If it does not exist, check the parent scope
        if self.parent:
            return self.parent.resolve(variable_name)
        # If it does not exist in the parent scope, raise an exception
        raise RuntimeError(f'Variable "{variable_name}" not found in scope')

    def setup_scope(self) -> None:
        """Setup the global scope."""
        # Declare global variables
        self.declare_variable("true", BooleanValue(True), True)
        self.declare_variable("false", BooleanValue(False), True)
        self.declare_variable("null", NullValue(), True)

        # Declare native methods
        self.declare_variable("print", NativeFunctionValue(_print), True)
        self.declare_variable("time", NativeFunctionValue(_time), True)
        self.declare_variable("input", NativeFunctionValue(_input), True)
        self.declare_variable("read", NativeFunctionValue(_readfile), True)
        self.declare_variable("len", NativeFunctionValue(_len), True)
        self.declare_variable("exit", NativeFunctionValue(_exit), True)
        self.declare_variable("str", NativeFunctionValue(_str), True)
        self.declare_variable("int", NativeFunctionValue(_int), True)
        self.declare_variable("float", NativeFunctionValue(_float), True)
        self.declare_variable("bool", NativeFunctionValue(_bool), True)
        self.declare_variable("array", NativeFunctionValue(_array), True)
        self.declare_variable("type", NativeFunctionValue(_type), True)
        self.declare_variable("sum", NativeFunctionValue(_sum), True)
        self.declare_variable("min", NativeFunctionValue(_min), True)
        self.declare_variable("max", NativeFunctionValue(_max), True)


def get_value(value: RuntimeValue, inside_array: bool = False) -> object:
    """Get the value of a RuntimeValue."""
    result = ""

    if isinstance(value, NullValue):
        result = None

    elif isinstance(value, (BooleanValue, NumberValue)):
        result = value.value

    elif isinstance(value, StringValue):
        if inside_array:
            result = '"' + value.value + '"'
        else:
            result = value.value

    elif isinstance(value, NativeFunctionValue):
        result = f"<native function {value.call.__name__[1:]}>"

    elif isinstance(value, FunctionValue):
        result = f"<function {value.name}>"

    elif isinstance(value, ArrayValue):
        result += "[ "
        for val in value.elements:
            result += f"{get_value(val, inside_array=True)}, "
        result = result[:-2] + " ]"

    elif isinstance(value, ObjectValue):
        result += "{ "
        for key, val in value.properties.items():
            result += f"{key}: {get_value(val, inside_array=True)}, "
        result = result[:-2] + " }"

    else:
        result = str(value)

    return result


# Native functions
def _print(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    values = []
    for arg in args:
        values.append(get_value(arg))
    print(*values)
    return NullValue()


def _time(_: list[RuntimeValue], __: Environment) -> RuntimeValue:
    return NumberValue(time.time())


def _input(args: list[RuntimeValue], __: Environment) -> RuntimeValue:
    result = input() if not args else input(args[0].value)
    return StringValue(result)


def _readfile(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    with open(args[0].value, "r", encoding="utf8") as file:
        return StringValue(file.read())


def _len(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if isinstance(args[0], StringValue):
        return NumberValue(len(args[0].value))

    if isinstance(args[0], ArrayValue):
        return NumberValue(len(args[0].elements))

    if isinstance(args[0], ObjectValue):
        return NumberValue(len(args[0].properties))

    raise RuntimeError(f"Cannot get length of {args[0]}")


def _exit(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    sys.exit(args[0].value)


def _str(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) == 0:
        return StringValue("")
    return StringValue(get_value(args[0]))


def _int(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) == 0:
        return NumberValue(0)
    if isinstance(args[0], (StringValue, NumberValue)):
        return NumberValue(int(args[0].value))
    raise RuntimeError(f"Cannot convert {args[0]} to int")


def _float(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) == 0:
        return NumberValue(0.0)
    if isinstance(args[0], (StringValue, NumberValue)):
        return NumberValue(float(args[0].value))
    raise RuntimeError(f"Cannot convert {args[0]} to float")


def _bool(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) == 0:
        return BooleanValue(False)
    if isinstance(args[0], (StringValue, NumberValue, BooleanValue)):
        return BooleanValue(bool(args[0].value))
    if isinstance(args[0], ArrayValue):
        return BooleanValue(bool(args[0].elements))
    if isinstance(args[0], ObjectValue):
        return BooleanValue(bool(args[0].properties))
    raise RuntimeError(f"Cannot convert {args[0]} to bool")


def _array(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    return ArrayValue(args)


def _type(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    return StringValue(type(args[0]).__name__)


def _sum(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) == 0:
        return NumberValue(0)
    if isinstance(args[0], ArrayValue):
        if all(isinstance(i, NumberValue) for i in args[0].elements):
            return NumberValue(sum(i.value for i in args[0].elements))
    raise RuntimeError(f"Cannot sum {args[0]}")


def _min(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) == 0:
        return NumberValue(0)
    if isinstance(args[0], ArrayValue):
        if all(isinstance(i, NumberValue) for i in args[0].elements):
            return NumberValue(min(i.value for i in args[0].elements))
    raise RuntimeError(f"Cannot get min for {args[0]}")


def _max(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) == 0:
        return NumberValue(0)
    if isinstance(args[0], ArrayValue):
        if all(isinstance(i, NumberValue) for i in args[0].elements):
            return NumberValue(max(i.value for i in args[0].elements))
    raise RuntimeError(f"Cannot get max for {args[0]}")
